@@ -26,7 +26,6 @@ import java.util.Locale;
 import java.util.Map;
 import java.util.concurrent.ScheduledFuture;
 import java.util.concurrent.ScheduledThreadPoolExecutor;
-import java.util.concurrent.ThreadFactory;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicBoolean;
 import java.util.concurrent.atomic.AtomicLong;
@@ -56,6 +55,7 @@ import org.apache.qpid.jms.provider.ProviderListener;
 import org.apache.qpid.jms.provider.ProviderRedirectedException;
 import org.apache.qpid.jms.provider.WrappedAsyncResult;
 import org.apache.qpid.jms.util.IOExceptionSupport;
+import org.apache.qpid.jms.util.QpidJMSThreadFactory;
 import org.apache.qpid.jms.util.ThreadPoolUtils;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
@@ -137,34 +137,14 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
     public FailoverProvider(List<URI> uris, Map<String, String> nestedOptions) {
         this.uris = new FailoverUriPool(uris, nestedOptions);
 
-        serializer = new ScheduledThreadPoolExecutor(1, new ThreadFactory() {
-
-            @Override
-            public Thread newThread(Runnable runner) {
-                Thread serial = new Thread(runner);
-                serial.setDaemon(true);
-                serial.setName("FailoverProvider: serialization thread");
-                return serial;
-            }
-        });
-
+        serializer = new ScheduledThreadPoolExecutor(1, new QpidJMSThreadFactory("FailoverProvider: serialization thread", true));
         serializer.setExecuteExistingDelayedTasksAfterShutdownPolicy(false);
         serializer.setContinueExistingPeriodicTasksAfterShutdownPolicy(false);
 
         // All Connection attempts happen in this schedulers thread.  Once a connection
         // is established it will hand the open connection back to the serializer thread
         // for state recovery.
-        connectionHub = new ScheduledThreadPoolExecutor(1, new ThreadFactory() {
-
-            @Override
-            public Thread newThread(Runnable runner) {
-                Thread serial = new Thread(runner);
-                serial.setDaemon(true);
-                serial.setName("FailoverProvider: connect thread");
-                return serial;
-            }
-        });
-
+        connectionHub = new ScheduledThreadPoolExecutor(1, new QpidJMSThreadFactory("FailoverProvider: connect thread", true));
         connectionHub.setExecuteExistingDelayedTasksAfterShutdownPolicy(false);
         connectionHub.setContinueExistingPeriodicTasksAfterShutdownPolicy(false);
     }
@@ -614,21 +594,31 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
                 try {
                     FailoverProvider.this.provider = provider;
                     provider.setProviderListener(FailoverProvider.this);
+                    connectedURI = provider.getRemoteURI();
 
                     if (!firstConnection) {
                         LOG.debug("Signalling connection recovery: {}", provider);
 
                         // Stage 1: Allow listener to recover its resources
-                        listener.onConnectionRecovery(provider);
+                        try {
+                            listener.onConnectionRecovery(provider);
+                        } finally {
+                            // Stage 2: If the provider knows of others lets add them to the URI pool
+                            //          even if something failed here we can learn of new hosts so we
+                            //          always process the potential Open frame failover URI results.
+                            processAlternates(provider.getAlternateURIs());
+                        }
 
-                        // Stage 2: Connection state recovered, get newly configured message factory.
+                        // Stage 3: Connection state recovered, get newly configured message factory.
                         FailoverProvider.this.messageFactory.set(provider.getMessageFactory());
 
-                        // Stage 3: Restart consumers, send pull commands, etc.
+                        // Stage 4: Restart consumers, send pull commands, etc.
                         listener.onConnectionRecovered(provider);
 
-                        // Stage 4: Let the client know that connection has restored.
+                        // Stage 5: Let the client know that connection has restored.
                         listener.onConnectionRestored(provider.getRemoteURI());
+                    } else {
+                        processAlternates(provider.getAlternateURIs());
                     }
 
                     // Last step: Send pending actions.
@@ -639,7 +629,6 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
 
                     nextReconnectDelay = reconnectDelay;
                     reconnectAttempts = 0;
-                    connectedURI = provider.getRemoteURI();
                     uris.connected();
 
                     // Cancel timeout processing since we are connected again.  We waited until
@@ -702,26 +691,33 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
 
                 reconnectAttempts++;
                 Throwable failure = null;
-                URI target = uris.getNext();
-                if (target != null) {
-                    Provider provider = null;
-                    try {
-                        LOG.debug("Connection attempt:[{}] to: {} in-progress", reconnectAttempts, target);
-                        provider = ProviderFactory.create(target);
-                        provider.connect(connectionInfo);
-                        initializeNewConnection(provider);
-                        return;
-                    } catch (Throwable e) {
-                        LOG.info("Connection attempt:[{}] to: {} failed", reconnectAttempts, target);
-                        failure = e;
+                if (!uris.isEmpty()) {
+                    for (int i = 0; i < uris.size(); ++i) {
+                        URI target = uris.getNext();
+                        if (target == null) {
+                            LOG.warn("Failover URI collection unexpectedly modified during connection attempt.");
+                            continue;
+                        }
+
+                        Provider provider = null;
                         try {
-                            if (provider != null) {
-                                provider.close();
-                            }
-                        } catch (Throwable ex) {}
+                            LOG.debug("Connection attempt:[{}] to: {} in-progress", reconnectAttempts, target);
+                            provider = ProviderFactory.create(target);
+                            provider.connect(connectionInfo);
+                            initializeNewConnection(provider);
+                            return;
+                        } catch (Throwable e) {
+                            LOG.info("Connection attempt:[{}] to: {} failed", reconnectAttempts, target);
+                            failure = e;
+                            try {
+                                if (provider != null) {
+                                    provider.close();
+                                }
+                            } catch (Throwable ex) {}
+                        }
                     }
                 } else {
-                    LOG.debug("No target URI available to connect to");
+                    LOG.debug("No remote URI available to connect to in failover list");
                 }
 
                 if (reconnectLimit != UNLIMITED && reconnectAttempts >= reconnectLimit) {
@@ -857,61 +853,42 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
         });
     }
 
-    @Override
-    public void onRemoteDiscovery(final List<URI> discovered) {
-        if (closingConnection.get() || closed.get() || failed.get()) {
-            return;
-        }
-
-        if (discovered == null || discovered.isEmpty()) {
-            return;
-        }
-
-        serializer.execute(new Runnable() {
-            @Override
-            public void run() {
-                if (!closingConnection.get() && !closed.get() && !failed.get()) {
-
-                    List<URI> newRemotes = new ArrayList<URI>(discovered);
-                    switch (amqpOpenServerListAction) {
-                        case ADD:
-                            try {
-                                uris.addAll(discovered);
-                            } catch (Throwable err) {
-                                LOG.warn("Error while attempting to add discovered URIs: {}", discovered);
-                            }
-                            break;
-                        case REPLACE:
-                            // The current server is assumed not to be in the list of updated remote
-                            // as it is meant for the failover nodes. The pool will de-dup if it is.
-                            newRemotes.add(0, connectedURI);
-                            try {
-                                uris.replaceAll(newRemotes);
-                            } catch (Throwable err) {
-                                LOG.warn("Error while attempting to add discovered URIs: {}", discovered);
-                            }
-                            break;
-                        case IGNORE:
-                            // Do Nothing
-                            break;
-                        default:
-                            // Shouldnt get here, but do nothing if we do.
-                            break;
+    private void processAlternates(List<URI> alternates) {
+        if (!alternates.isEmpty()) {
+            List<URI> newRemotes = new ArrayList<URI>(alternates);
+            switch (amqpOpenServerListAction) {
+                case ADD:
+                    try {
+                        uris.addAll(alternates);
+                    } catch (Throwable err) {
+                        LOG.warn("Error while attempting to add discovered URIs: {}", alternates);
                     }
-
-                    // Inform any listener that we've made a new discovery.
-                    if (listener != null) {
-                        listener.onRemoteDiscovery(discovered);
+                    break;
+                case REPLACE:
+                    // The current server is assumed not to be in the list of updated remote
+                    // as it is meant for the failover nodes. The pool will de-dup if it is.
+                    newRemotes.add(0, connectedURI);
+                    try {
+                        LOG.debug("Replacing uris:{} with new set: {}", uris, newRemotes);
+                        uris.replaceAll(newRemotes);
+                    } catch (Throwable err) {
+                        LOG.warn("Error while attempting to add discovered URIs: {}", alternates);
                     }
-                }
+                    break;
+                case IGNORE:
+                    // Do Nothing
+                    break;
+                default:
+                    // Shouldn't get here, but do nothing if we do.
+                    break;
             }
-        });
+        }
     }
 
     //--------------- URI update and rebalance methods -----------------------//
 
     public void add(final URI uri) {
-        serializer.execute(new Runnable() {
+        connectionHub.execute(new Runnable() {
             @Override
             public void run() {
                 uris.add(uri);
@@ -920,7 +897,7 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
     }
 
     public void remove(final URI uri) {
-        serializer.execute(new Runnable() {
+        connectionHub.execute(new Runnable() {
             @Override
             public void run() {
                 uris.remove(uri);
@@ -939,6 +916,16 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
         return null;
     }
 
+
+    @Override
+    public List<URI> getAlternateURIs() {
+        Provider provider = this.provider;
+        if (provider != null) {
+            return provider.getAlternateURIs();
+        }
+        return null;
+    };
+
     @Override
     public void setProviderListener(ProviderListener listener) {
         this.listener = listener;
@@ -1248,6 +1235,7 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
                     if (firstConnection) {
                         LOG.trace("First connection requst has completed:");
                         FailoverProvider.this.messageFactory.set(provider.getMessageFactory());
+                        processAlternates(provider.getAlternateURIs());
                         listener.onConnectionEstablished(provider.getRemoteURI());
                         firstConnection = false;
                     } else {
@@ -1269,6 +1257,9 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
                 serializer.execute(new Runnable() {
                     @Override
                     public void run() {
+                        // If we managed to receive an Open frame it might contain
+                        // a failover update so process it before handling the error.
+                        processAlternates(provider.getAlternateURIs());
                         handleProviderFailure(IOExceptionSupport.create(result));
                     }
                 });
@@ -1282,5 +1273,5 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
 
     private static enum FailoverServerListAction {
         ADD, REPLACE, IGNORE
-    };
+    }
 }
